# flake8: noqa
import pprint

import pytest

import adijif


@pytest.mark.parametrize("attr", ["data_path_width", "K", "F", "L", "M", "N", "Np"])
def test_jesd_ints(attr):
    with pytest.raises(Exception, match=f"{attr} must be an integer"):
        cnv = adijif.ad9680()
        setattr(cnv, attr, 0.5)


@pytest.mark.parametrize("attr", ["K", "F", "L", "M", "N", "Np"])
def test_jesd_oor(attr):
    with pytest.raises(Exception, match=f"{attr} not in range for device"):
        cnv = adijif.ad9680()
        setattr(cnv, attr, 1024)



# flake8: noqa

import pytest

import adijif


def test_converter_lane_count_valid():
    sys = adijif.system("ad9144", "ad9523_1", "xilinx", 125e6)
    sys.fpga.setup_by_dev_kit_name("zcu102")

    sys.converter.sample_clock = 1e9
    # Mode 0
    sys.converter.interpolation = 1
    sys.converter.L = 8
    sys.converter.M = 4
    sys.converter.N = 16
    sys.converter.Np = 16
    sys.converter.K = 32
    sys.converter.F = 1
    sys.converter.HD = 1
    sys.converter.clocking_option = "integrated_pll"

    cfg = sys.solve()


def test_converter_lane_count_exceeds_fpga_lane_count():
    convs = ["ad9144", "ad9144", "ad9144"]
    sys = adijif.system(convs, "ad9523_1", "xilinx", 125e6)
    sys.fpga.setup_by_dev_kit_name("zcu102")

    for i, _ in enumerate(convs):
        # Mode 0
        sys.converter[i].sample_clock = 1e9
        sys.converter[i].interpolation = 1
        sys.converter[i].L = 8
        sys.converter[i].M = 4
        sys.converter[i].N = 16
        sys.converter[i].Np = 16
        sys.converter[i].K = 32
        sys.converter[i].F = 1
        sys.converter[i].HD = 1
        sys.converter[i].clocking_option = "integrated_pll"

    with pytest.raises(Exception, match=f"Max SERDES lanes exceeded. 8 only available"):
        cfg = sys.solve()


def test_nested_converter_lane_count_valid():
    sys = adijif.system("adrv9009", "ad9528", "xilinx", 122.88e6)
    sys.fpga.setup_by_dev_kit_name("zcu102")
    sys.converter.adc.sample_clock = 122.88e6
    sys.converter.dac.sample_clock = 122.88e6

    sys.converter.adc.decimation = 4
    sys.converter.dac.interpolation = 4

    mode_rx = "17"
    mode_tx = "6"
    sys.converter.adc.set_quick_configuration_mode(mode_rx, "jesd204b")
    sys.converter.dac.set_quick_configuration_mode(mode_tx, "jesd204b")

    cfg = sys.solve()


def test_nested_converter_lane_count_exceeds_fpga_lane_count():
    fpga_L = 1

    sys = adijif.system("adrv9009", "ad9528", "xilinx", 122.88e6)

    sys.fpga.setup_by_dev_kit_name("zcu102")
    sys.fpga.max_serdes_lanes = fpga_L  # Force it to break

    sys.converter.adc.sample_clock = 122.88e6
    sys.converter.dac.sample_clock = 122.88e6

    sys.converter.adc.decimation = 4
    sys.converter.dac.interpolation = 4

    mode_rx = "17"
    mode_tx = "6"
    sys.converter.adc.set_quick_configuration_mode(mode_rx, "jesd204b")
    sys.converter.dac.set_quick_configuration_mode(mode_tx, "jesd204b")

    with pytest.raises(
        Exception, match=f"Max SERDES lanes exceeded. {fpga_L} only available"
    ):
        cfg = sys.solve()



# flake8: noqa
import pprint

import adijif as jif


def test_adf4030_sysref_example():
    ref_in = int(100e6)

    sync = jif.adf4030()

    output_clocks = [
        2.5e6,
    ]
    output_clocks = list(map(int, output_clocks))  # force to be ints
    clock_names = [
        "SYSREF",
    ]

    sync.set_requested_clocks(ref_in, output_clocks, clock_names)

    sync.solve()

    o = sync.get_config()

    pprint.pprint(o)

    assert o["output_clocks"]["SYSREF"]["rate"] == output_clocks[0]
    assert o["output_clocks"]["SYSREF"]["divider"] == 992


def test_adf4030_system():
    vcxo = 100e6

    import adijif

    sys = adijif.system("ad9081", "hmc7044", "xilinx", vcxo, solver="CPLEX")
    sys.fpga.setup_by_dev_kit_name("zcu102")
    sys.fpga.ref_clock_constraint = "Unconstrained"
    sys.fpga.sys_clk_select = "SYSCLK_QPLL0"  # Use faster QPLL
    sys.fpga.out_clk_select = (
        "XCVR_PROGDIV_CLK"  # force reference to be core clock rate
    )
    cddc = 6
    fddc = 4
    sys.converter.adc.sample_clock = 2900000000 / (cddc * fddc)
    sys.converter.dac.sample_clock = 5800000000 / (cddc * fddc)
    # sys.converter.adc.decimation = 8 * 6
    # sys.converter.dac.interpolation = 4 * 12
    sys.converter.adc.datapath.cddc_decimations = [cddc] * 4
    sys.converter.adc.datapath.fddc_decimations = [fddc] * 8
    sys.converter.adc.datapath.fddc_enabled = [True] * 8
    sys.converter.dac.datapath.cduc_interpolation = cddc
    sys.converter.dac.datapath.fduc_interpolation = fddc
    sys.converter.dac.datapath.fduc_enabled = [True] * 8
    assert sys.converter.dac.interpolation == cddc * fddc

    # sys.converter.clocking_option = "direct"
    # sys.add_pll_inline("adf4371", sys.clock, sys.converter)
    sys.add_pll_sysref("adf4030", vcxo, sys.converter, sys.fpga)

    mode_tx = "0"
    mode_rx = "1.0"

    sys.converter.dac.set_quick_configuration_mode(mode_tx, "jesd204c")
    sys.converter.adc.set_quick_configuration_mode(mode_rx, "jesd204c")

    assert sys.converter.adc.M == 8
    assert sys.converter.adc.F == 12
    assert sys.converter.adc.K == 64
    assert sys.converter.adc.Np == 12
    assert sys.converter.adc.CS == 0
    assert sys.converter.adc.L == 1
    assert sys.converter.adc.S == 1

    sys.converter.adc._check_clock_relations()
    sys.converter.dac._check_clock_relations()

    cfg = sys.solve()

    pprint.pprint(cfg)
    print(sys.converter.dac.converter_clock)



# flake8: noqa
import pprint

import pytest

import adijif as jif

# def test_generate_max_rates_fpga_limits_utility():
#     conv = jif.ad9081_rx()

#     # Determine max data rate from FPGA
#     fpga = jif.xilinx()
#     fpga.setup_by_dev_kit_name("zc706")
#     fpga.sys_clk_select = "GTH34_SYSCLK_QPLL0"  # Use faster QPLL

#     limits = {"jesd_class": "jesd204b", "bit_clock": {"<=": 10e9}}

#     results = jif.utils.get_max_sample_rates(conv, fpga, limits)

#     ref = []
#     assert results == ref


def test_generate_max_rates_fpga_utility():
    conv = jif.ad9081_rx()
    # conv.decimation = 1

    # Determine max data rate from FPGA
    fpga = jif.xilinx()
    fpga.setup_by_dev_kit_name("zc706")
    fpga.sys_clk_select = "XCVR_QPLL0"  # Use faster QPLL

    limits = {"jesd_class": "jesd204b"}

    results = jif.utils.get_max_sample_rates(conv, fpga, limits)

    ref = [
        {
            "sample_clock": 4000000000.0,
            "bit_clock": 10000000000.0,
            "L": 8,
            "M": 1,
            "quick_configuration_mode": "19.01",
            "jesd_class": "jesd204b",
        },
        {
            "sample_clock": 2062500000.0,
            "bit_clock": 10312500000.0,
            "L": 8,
            "M": 2,
            "quick_configuration_mode": "19.0",
            "jesd_class": "jesd204b",
        },
        {
            "sample_clock": 515625000.0,
            "bit_clock": 10312500000.0,
            "L": 3,
            "M": 3,
            "quick_configuration_mode": "9.01",
            "jesd_class": "jesd204b",
        },
        {
            "sample_clock": 1031250000.0,
            "bit_clock": 10312500000.0,
            "L": 8,
            "M": 4,
            "quick_configuration_mode": "18.0",
            "jesd_class": "jesd204b",
        },
        {
            "sample_clock": 515625000.0,
            "bit_clock": 10312500000.0,
            "L": 6,
            "M": 6,
            "quick_configuration_mode": "15.01",
            "jesd_class": "jesd204b",
        },
        {
            "sample_clock": 515625000.0,
            "bit_clock": 10312500000.0,
            "L": 8,
            "M": 8,
            "quick_configuration_mode": "16.0",
            "jesd_class": "jesd204b",
        },
        {
            "sample_clock": 257812500.0,
            "bit_clock": 10312500000.0,
            "L": 6,
            "M": 12,
            "quick_configuration_mode": "15.1",
            "jesd_class": "jesd204b",
        },
        {
            "sample_clock": 257812500.0,
            "bit_clock": 10312500000.0,
            "L": 8,
            "M": 16,
            "quick_configuration_mode": "17.0",
            "jesd_class": "jesd204b",
        },
    ]
    assert len(results) == len(ref)
    for result in results:
        # print(result)
        if result not in ref:
            pprint.pprint(result)
        assert result in ref


def test_generate_max_rates_utility():
    conv = jif.ad9081_rx()
    # conv.decimation = 1

    results = jif.utils.get_max_sample_rates(conv)

    ref = [
        {
            "L": 8,
            "M": 1,
            "bit_clock": 10000000000.0,
            "quick_configuration_mode": "19.01",
            "sample_clock": 4000000000.0,
            "jesd_class": "jesd204b",
        },
        {
            "L": 4,
            "M": 2,
            "bit_clock": 24750000000.0,
            "quick_configuration_mode": "14.0",
            "sample_clock": 4000000000.0,
            "jesd_class": "jesd204c",
        },
        {
            "L": 3,
            "M": 3,
            "bit_clock": 24750000000.0,
            "quick_configuration_mode": "9.01",
            "sample_clock": 1500000000.0,
            "jesd_class": "jesd204c",
        },
        {
            "L": 8,
            "M": 4,
            "bit_clock": 24750000000.0,
            "quick_configuration_mode": "27.0",
            "sample_clock": 4000000000.0,
            "jesd_class": "jesd204c",
        },
        {
            "L": 6,
            "M": 6,
            "bit_clock": 24750000000.0,
            "quick_configuration_mode": "15.01",
            "sample_clock": 1500000000.0,
            "jesd_class": "jesd204c",
        },
        {
            "L": 8,
            "M": 8,
            "bit_clock": 24750000000.0,
            "quick_configuration_mode": "26.0",
            "sample_clock": 2000000000.0,
            "jesd_class": "jesd204c",
        },
        {
            "L": 6,
            "M": 12,
            "bit_clock": 24750000000.0,
            "quick_configuration_mode": "15.0",
            "sample_clock": 750000000.0,
            "jesd_class": "jesd204c",
        },
        {
            "L": 8,
            "M": 16,
            "bit_clock": 24750000000.0,
            "quick_configuration_mode": "17.0",
            "sample_clock": 750000000.0,
            "jesd_class": "jesd204c",
        },
    ]

    # pprint.pprint(results)

    assert len(results) == len(ref)
    for result in results:
        if result not in ref:
            pprint.pprint(result)
        assert result in ref



# flake8: noqa
"""Bindings for the D2 Compiler"""

import ctypes
import os

loc = os.path.dirname(os.path.abspath(__file__))
lib_path = os.path.join(loc, "d2lib.so")

if not os.path.exists(lib_path):
    raise FileNotFoundError(f"Could not find {lib_path}")

library = ctypes.cdll.LoadLibrary(lib_path)

runme = library.runme
runme.argtypes = [ctypes.c_char_p]
runme.restype = ctypes.c_char_p


def compile(code: str) -> str:
    try:
        graph_bytes = runme(code.encode("utf-8"))
        return graph_bytes.decode("utf-8")
    except Exception as e:
        print(e)
        return None



# flake8: noqa
import os
import pprint

import pytest

import adijif
import adijif.converters
import adijif.converters.ad9084_util

here = os.path.dirname(os.path.abspath(__file__))
profile_json = os.path.join(
    here, "apollo_profiles", "ad9084_profiles", "id00_stock_mode.json"
)
summary_json = os.path.join(
    here, "apollo_profiles", "ad9084_profiles", "id00_stock_mode.summary"
)


ref = {
    "common_lane_rate_Hz": 20625000000,
    "core_clock_Hz": None,
    "datapath": {
        "cddc_decimation": 4,
        "cduc_interpolation": 4,
        "fddc_decimation": 2,
        "fduc_interpolation": 2,
    },
    "device_clock_Hz": 20000000000,
    "id": "id00_stock_mode",
    "is_8t8r": False,
    "jesd_settings": {
        "jrx": {
            "F": 1,
            "HD": True,
            # "K": 256,
            "L": 8,
            "M": 4,
            # "N": 16,
            "Np": 16,
            "S": 1,
        },
        "jtx": {
            "F": 1,
            "HD": True,
            # "K": 256,
            "L": 8,
            "M": 4,
            # "N": 16,
            "Np": 16,
            "S": 1,
        },
    },
    "profile_name": "id00_stock_mode",
    "rx_jesd_mode": 47,
    "tx_jesd_mode": 47,
}


def test_config_parser():

    settings = adijif.converters.ad9084_util.parse_json_config(profile_json)

    print("Settings:")
    pprint.pprint(settings)

    assert settings is not None

    # Verify the settings match the reference
    for key in ref:
        if isinstance(ref[key], dict):
            assert settings[key] == ref[key]
        else:
            assert settings[key] == ref[key], f"Mismatch for key: {key}"


def test_ad9084_separate_model_update():

    settings = adijif.converters.ad9084_util.parse_json_config(profile_json)

    print("Settings:")
    pprint.pprint(settings)

    assert settings is not None

    # Apply the settings to the AD9084 RX
    vcxo = 125000000
    clock = "hmc7044"
    sys = adijif.system("ad9084_rx", clock, "xilinx", vcxo, solver="CPLEX")

    adijif.converters.ad9084_util.apply_settings(sys.converter, settings)

    # Verify the settings were applied correctly
    # converter_rate / (cddc_dec * fddc_dec)
    assert sys.converter.converter_clock == settings["device_clock_Hz"]
    assert sys.converter.sample_clock == settings["device_clock_Hz"] / (
        settings["datapath"]["cddc_decimation"]
        * settings["datapath"]["fddc_decimation"]
    )
    # Check the JESD settings
    keys = settings["jesd_settings"]["jrx"].keys()
    for key in keys:
        assert (
            getattr(sys.converter, key) == settings["jesd_settings"]["jtx"][key]
        ), f"Mismatch for key: {key}"


def test_ad9084_model_update():

    settings = adijif.converters.ad9084_util.parse_json_config(profile_json)

    print("Settings:")
    pprint.pprint(settings)

    assert settings is not None

    # Apply the settings to the AD9084 RX
    vcxo = 125000000
    clock = "hmc7044"
    sys = adijif.system("ad9084_rx", clock, "xilinx", vcxo, solver="CPLEX")

    sys.converter.apply_profile_settings(profile_json)

    # Verify the settings were applied correctly
    # converter_rate / (cddc_dec * fddc_dec)
    assert sys.converter.converter_clock == settings["device_clock_Hz"]
    assert sys.converter.sample_clock == settings["device_clock_Hz"] / (
        settings["datapath"]["cddc_decimation"]
        * settings["datapath"]["fddc_decimation"]
    )
    # Check the JESD settings
    keys = settings["jesd_settings"]["jrx"].keys()
    for key in keys:
        assert (
            getattr(sys.converter, key) == settings["jesd_settings"]["jtx"][key]
        ), f"Mismatch for key: {key}"



"""ADI JIF utility types and functions."""

from typing import Dict, Union

import numpy as np

from adijif.solvers import GEKKO, CpoModel, integer_var  # type: ignore


class range:
    """Model range of possible values.

    This class is similar to range but interfaces with
    the modeling layers to create correct mappings for internal
    solvers
    """

    def __init__(self, start: int, stop: int, step: int, name: str) -> None:
        """Model range of possible values.

        This class is similar to range but interfaces with
        the modeling layers to create correct mappings for internal
        solvers

        Args:
            start (int): Start value of range
            stop (int): Last value of range-step
            step (int): Difference between successive values in range
            name (str): Name of variable
        """
        assert isinstance(start, int), "start must be an int"
        assert isinstance(stop, int), "stop must be an int"
        assert isinstance(step, int), "step must be an int"
        assert isinstance(name, str), "name must be a string"

        self.start = start
        self.stop = stop
        self.step = step
        self.name = name

    def __call__(self, model: Union[GEKKO, CpoModel]) -> Dict:
        """Generate range for parameter solver.

        Args:
            model (GEKKO, CpoModel): Model of JESD system or part to solve

        Returns:
            Dict: Dictionary of solver variable(s) for model
        """
        if GEKKO and CpoModel:
            assert isinstance(
                model, (GEKKO, CpoModel)
            ), "range must be called with input type model"
        elif GEKKO:
            assert isinstance(
                model, GEKKO
            ), "range must be called with input type model"
        elif CpoModel:
            assert isinstance(
                model, CpoModel
            ), "range must be called with input type model"

        config = {}
        if isinstance(model, CpoModel):
            o_array = np.arange(self.start, self.stop, self.step)
            config["range"] = integer_var(domain=o_array, name=self.name + "_Var")
            return config

        if self.step == 1:
            config["range"] = model.Var(
                integer=True,
                lb=self.start,
                ub=self.stop,
                value=self.start,
                name=self.name + "_Var",
            )
        else:
            # Create piecewise problem with binary values (DO NOT USE SOS1!)
            o_array = np.arange(self.start, self.stop, self.step)
            o_array = list(map(int, o_array))

            options = model.Array(model.Var, len(o_array), lb=0, ub=1, integer=True)
            model.Equation(model.sum(options) == 1)  # only select one
            config["range"] = model.Intermediate(model.sum(o_array * options))

        return config


class arb_source:
    """Arbitrary source for solver.

    This internally uses a division of two integers to create an arbitrary clock
    source.
    """

    _max_scalar = int(1e11)

    def __init__(self, name: str) -> None:
        """Arbitrary source for solver.

        Args:
            name (str): Name of source
        """
        self.name = name

        self._a = integer_var(0, self._max_scalar, name=name + "_a")
        self._b = integer_var(0, self._max_scalar, name=name + "_b")

    def __call__(self, model: Union[GEKKO, CpoModel]) -> Dict:
        """Generate arbitrary source for solver.

        Args:
            model (GEKKO, CpoModel): Model of JESD system or part to solve

        Returns:
            Dict: Dictionary of solver variable(s) for model

        Raises:
            NotImplementedError: Only CpoModel is supported
        """
        if GEKKO and CpoModel:
            assert isinstance(
                model, (GEKKO, CpoModel)
            ), "arb_source must be called with input type model"
        elif GEKKO:
            assert isinstance(
                model, GEKKO
            ), "arb_source must be called with input type model"
        elif CpoModel:
            assert isinstance(
                model, CpoModel
            ), "arb_source must be called with input type model"

        # config = {}
        if isinstance(model, CpoModel):
            # config[self.name] = self._a / self._b
            # return config
            return self._a / self._b

        raise NotImplementedError("Only CpoModel is supported")

    def get_config(self, solution: Dict) -> Dict:
        """Get configuration from solver results.

        Args:
            solution (Dict): Solver results

        Returns:
            Dict: Dictionary of solver variable(s) for model
        """
        return {self.name: solution[self._a] / solution[self._b]}



# flake8: noqa
import pprint

import pytest

import adijif


def test_jesd_unknown_dev():
    name = "zcu103"
    with pytest.raises(Exception, match=f"No boardname found in library for {name}"):
        f = adijif.xilinx()
        f.setup_by_dev_kit_name(name)


def test_jesd_wrong_order():
    msg = (
        "get_required_clocks must be run to generated"
        + " dependent clocks before names are available"
    )
    with pytest.raises(Exception, match=msg):
        f = adijif.xilinx()
        f.ref_clock_min = 60000000
        f.ref_clock_max = 670000000
        cn = f.get_required_clock_names()


def test_ref_clock_not_set():
    msg = "ref_clock_min or ref_clock_max not set"
    with pytest.raises(Exception, match=msg):
        f = adijif.xilinx()
        cn = f.get_required_clocks([], [])



"""LTC6952 clock chip model."""

from typing import Dict, List, Union

from docplex.cp.solution import CpoSolveResult  # type: ignore

from adijif.clocks.ltc6952_bf import ltc6952_bf
from adijif.solvers import CpoExpr, GK_Intermediate


class ltc6952(ltc6952_bf):
    """LTC6952 clock chip model.

    This model currently supports VCXO+PLL2 configurations
    """

    name = "LTC6952"

    vcxo = 125000000

    # Ranges
    r2_divider_min = 1
    r2_divider_max = 1023
    r2_available = [*range(1, 1023 + 1)]
    n2_divider_min = 1
    n2_divider_max = 65535
    n2_available = [*range(1, 65535 + 1)]

    """ Output dividers """
    d_available = [
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
        13,
        14,
        15,
        16,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
        24,
        25,
        26,
        27,
        28,
        29,
        30,
        31,
        32,
        34,
        36,
        38,
        40,
        42,
        44,
        46,
        48,
        50,
        52,
        54,
        56,
        58,
        60,
        62,
        64,
        68,
        72,
        76,
        80,
        84,
        88,
        92,
        96,
        100,
        104,
        108,
        112,
        116,
        120,
        124,
        128,
        136,
        144,
        152,
        160,
        168,
        176,
        184,
        192,
        200,
        208,
        216,
        224,
        232,
        240,
        248,
        256,
        272,
        288,
        304,
        320,
        336,
        352,
        368,
        384,
        400,
        416,
        432,
        448,
        464,
        480,
        496,
        512,
        544,
        576,
        608,
        640,
        672,
        704,
        736,
        768,
        800,
        832,
        864,
        896,
        928,
        960,
        992,
        1024,
        1088,
        1152,
        1216,
        1280,
        1344,
        1408,
        1472,
        1536,
        1600,
        1664,
        1728,
        1792,
        1856,
        1920,
        1984,
        2048,
        2176,
        2304,
        2432,
        2560,
        2688,
        2816,
        2944,
        3072,
        3200,
        3328,
        3456,
        3584,
        3712,
        3840,
        3968,
        4096,
    ]

    # Defaults
    _d: Union[int, List[int]] = [
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
        13,
        14,
        15,
        16,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
        24,
        25,
        26,
        27,
        28,
        29,
        30,
        31,
        32,
        34,
        36,
        38,
        40,
        42,
        44,
        46,
        48,
        50,
        52,
        54,
        56,
        58,
        60,
        62,
        64,
        68,
        72,
        76,
        80,
        84,
        88,
        92,
        96,
        100,
        104,
        108,
        112,
        116,
        120,
        124,
        128,
        136,
        144,
        152,
        160,
        168,
        176,
        184,
        192,
        200,
        208,
        216,
        224,
        232,
        240,
        248,
        256,
        272,
        288,
        304,
        320,
        336,
        352,
        368,
        384,
        400,
        416,
        432,
        448,
        464,
        480,
        496,
        512,
        544,
        576,
        608,
        640,
        672,
        704,
        736,
        768,
        800,
        832,
        864,
        896,
        928,
        960,
        992,
        1024,
        1088,
        1152,
        1216,
        1280,
        1344,
        1408,
        1472,
        1536,
        1600,
        1664,
        1728,
        1792,
        1856,
        1920,
        1984,
        2048,
        2176,
        2304,
        2432,
        2560,
        2688,
        2816,
        2944,
        3072,
        3200,
        3328,
        3456,
        3584,
        3712,
        3840,
        3968,
        4096,
    ]
    _n2: Union[int, List[int]] = [*range(1, 65535 + 1)]
    _r2: Union[int, List[int]] = [*range(1, 1023 + 1)]

    # Limits
    """ Internal limits """
    _vco_min = 1e6  # uses external VCO limits need to be set by the user
    _vco_max = 4500e6
    pfd_max = 167e6
    vcxo_min = 1e6
    vcxo_max = 500e6

    minimize_feedback_dividers = False

    # State management
    _clk_names: List[str] = []

    @property
    def vco_min(self) -> float:
        """Actual lower VCO frequency.

        Valid range 1->4500 MHz

        Returns:
            float: Current vco minimum value
        """
        return self._vco_min

    @vco_min.setter
    def vco_min(self, value: float) -> None:
        """Actual lower VCO frequency.

                Valid range 1->4500 MHz

        Args:
            value (float): Allowable values for vco min

        """
        self._vco_min = value

    @property
    def vco_max(self) -> float:
        """Actual upper VCO frequency.

        Valid range 1->4500 MHz

        Returns:
            float: Current vco minimum value
        """
        return self._vco_max

    @vco_max.setter
    def vco_max(self, value: float) -> None:
        """Actual upper VCO frequency.

                Valid range 1->4500 MHz

        Args:
            value (float): Allowable values for vco min

        """
        self._vco_max = value

    @property
    def d(self) -> Union[int, List[int]]:
        """Output dividers.

        Valid dividers are 1,2,3,4,5,6..32->(even)->4096

        Returns:
            int: Current allowable dividers
        """
        return self._d

    @d.setter
    def d(self, value: Union[int, List[int]]) -> None:
        """Output dividers.

        Valid dividers are 1,2,3,4,5,6..32->(even)->4096

        Args:
            value (int, list[int]): Allowable values for divider

        """
        self._check_in_range(value, self.d_available, "d")
        self._d = value

    @property
    def n2(self) -> Union[int, List[int]]:
        """n2: VCO feedback divider.

        Valid dividers are 1->65536

        Returns:
            int: Current allowable dividers
        """
        return self._n2

    @n2.setter
    def n2(self, value: Union[int, List[int]]) -> None:
        """VCO feedback divider.

        Valid dividers are 1->65536

        Args:
            value (int, list[int]): Allowable values for divider

        """
        self._check_in_range(value, self.n2_available, "n2")
        self._n2 = value

    @property
    def r2(self) -> Union[int, List[int]]:
        """VCXO input dividers.

        Valid dividers are 1->4096

        Returns:
            int: Current allowable dividers
        """
        return self._r2

    @r2.setter
    def r2(self, value: Union[int, List[int]]) -> None:
        """VCXO input dividers.

        Valid dividers are 1->4096

        Args:
            value (int, list[int]): Allowable values for divider

        """
        self._check_in_range(value, self.r2_available, "r2")
        self._r2 = value

    def get_config(self, solution: CpoSolveResult = None) -> Dict:
        """Extract configurations from solver results.

        Collect internal clock chip configuration and output clock definitions
        leading to connected devices (converters, FPGAs)

        Args:
            solution (CpoSolveResult): CPlex solution. Only needed for CPlex solver

        Returns:
            Dict: Dictionary of clocking rates and dividers for configuration

        Raises:
            Exception: If solver is not called first
        """
        if not self._clk_names:
            raise Exception("set_requested_clocks must be called before get_config")

        if solution:
            self.solution = solution

        out_dividers = [self._get_val(x) for x in self.config["out_dividers"]]

        clk: float = (
            self.vcxo  # type: ignore # noqa: B950
            * self._get_val(self.config["n2"])  # type: ignore # noqa: B950
            / self._get_val(self.config["r2"])  # type: ignore # noqa: B950
        )

        config: Dict = {
            "r2": self._get_val(self.config["r2"]),
            "n2": self._get_val(self.config["n2"]),
            "VCO": clk,
            "vcxo": self.vcxo,
            "out_dividers": out_dividers,
            "output_clocks": [],
        }

        output_cfg = {}
        for i, div in enumerate(out_dividers):
            rate = clk / div  # type: ignore # noqa: B950
            output_cfg[self._clk_names[i]] = {"rate": rate, "divider": div}

        config["output_clocks"] = output_cfg

        self._saved_solution = config

        return config

    def _setup_solver_constraints(self, vcxo: int) -> None:
        """Apply constraints to solver model.

        Args:
            vcxo (int): VCXO frequency in hertz
        """
        self.vcxo = vcxo
        self.config = {
            "r2": self._convert_input(self._r2, "r2"),
            "n2": self._convert_input(self._n2, "n2"),
        }
        # self.config = {"r2": self.model.Var(integer=True, lb=1, ub=4095, value=1)}
        # self.config["n2"] = self.model.Var(
        #     integer=True, lb=8, ub=4095
        # )  # FIXME: CHECK UB

        # PLL2 equations
        self._add_equation(
            [
                vcxo / self.config["r2"] <= self.pfd_max,
                vcxo / self.config["r2"] * self.config["n2"] <= self.vco_max,
                vcxo / self.config["r2"] * self.config["n2"] >= self.vco_min,
            ]
        )

        # Objectives
        if self.minimize_feedback_dividers:
            self.model.minimize(self.config["r2"])
            # self.model.Obj(self.config["r2"])

    def _setup(self, vcxo: int) -> None:
        # Setup clock chip internal constraints

        # FIXME: ADD SPLIT m1 configuration support

        # Setup clock chip internal constraints
        self._setup_solver_constraints(vcxo)

        # Add requested clocks to output constraints
        self.config["out_dividers"] = []
        self._clk_names = []  # Reset

    def _get_clock_constraint(
        self, clk_name: List[str]
    ) -> Union[int, float, CpoExpr, GK_Intermediate]:
        """Get abstract clock output.

        Args:
            clk_name (str):  String of clock name

        Returns:
            (int or float or CpoExpr or GK_Intermediate): Abstract
                or concrete clock reference

        Raises:
            Exception: Invalid solver
        """
        if self.solver == "gekko":
            __d = self._d if isinstance(self._d, list) else [self._d]

            if __d.sort() != self.d_available.sort():
                raise Exception("For solver gekko d is not configurable for LTC6952")
            # Since d is so disjoint it is very annoying to solve.
            mp = self.model.Var(integer=True, lb=1, ub=32)
            nx = self.model.Var(integer=True, lb=0, ub=7)
            od = self.model.Intermediate(mp * pow(2, nx))
        elif self.solver == "CPLEX":
            od = self._convert_input(self._d, "d_" + str(clk_name))
        else:
            raise Exception("Unknown solver {}".format(self.solver))

        self.config["out_dividers"].append(od)
        self._clk_names.append(clk_name)
        return self.vcxo / self.config["r2"] * self.config["n2"] / od

    def set_requested_clocks(
        self, vcxo: int, out_freqs: List, clk_names: List[str]
    ) -> None:
        """Define necessary clocks to be generated in model.

        Args:
            vcxo (int): VCXO frequency in hertz
            out_freqs (List): list of required clocks to be output
            clk_names (List[str]):  list of strings of clock names

        Raises:
            Exception: If len(out_freqs) != len(clk_names)
        """
        if len(clk_names) != len(out_freqs):
            raise Exception("clk_names is not the same size as out_freqs")

        # Setup clock chip internal constraints
        self._setup(vcxo)
        self._clk_names = clk_names

        # Add requested clocks to output constraints
        for out_freq in out_freqs:
            if self.solver == "gekko":
                __d = self._d if isinstance(self._d, list) else [self._d]
                if __d.sort() != self.d_available.sort():
                    raise Exception(
                        "For solver gekko d is not configurable for LTC6952"
                    )

                mp = self.model.Var(integer=True, lb=1, ub=32)
                nx = self.model.Var(integer=True, lb=0, ub=7)
                od = self.model.Intermediate(mp * pow(2, nx))

            elif self.solver == "CPLEX":
                od = self._convert_input(self._d, "d_" + str(out_freq))

            self._add_equation(
                [self.vcxo / self.config["r2"] * self.config["n2"] / od == out_freq]
            )
            self.config["out_dividers"].append(od)

            # Objectives
            # self.model.Obj(-1*eo) # Favor even dividers
